#! /np/pub/bin/perl
#
# ---------------------------------------------------------------
# |comproc.pl - Script to read in an observation file, compute  |
# |             area under ROC curve and test for significant   |
# |             differences in the presence of covariates.      |
# |								|
# |Format of input file: 1.) Header line is optional		|
# |                      2.) 5 columns in the following order:	|
# |                          Name T1 T2 X D			|
# |                          Where: Name = observation name	|
# |                                   T1 = 1, 2,..., K		|
# |                                   T2 = 1, 2,..., K		|
# |                                    X = 1, 2,..., I		|
# |				       D = 0,1 or . for missing	| 
# |                      3.) File is space delimeted		|
# |								|
# |Format of table input file if -t option specified            |
# | 1.) Header line is optional                                 |
# | 2.) 5 columns in the following order:                       |
# |     T1 T2 X D COUNT                                         |
# |     T1 = 1, 2,..., K		                        |
# |     T2 = 1, 2,..., K		                        |
# |      X = 1, 2,..., I		                        |
# |      D = 0,1 or . for missing                       	| 
# |     COUNT = number of observations in cell                  |
# | 3.) File is space delimeted		                        |
# |                                                             |
# |If verbose mode (-v option) is specified the following files |
# |are created:                                                 |
# |                                                             |
# | INFILE.logitdat     - Data used for logistic regression     |
# | INFILE-logit.sas    - Logistic regr. sas file               |
# | INFILE-logit.lst    - Logistic regr. sas lst file           |
# | INFILE-logit.log    - Logistic regr. sas log file           |
# | INFILE-logit.beta   - Beta estimates & covariance matrix    |
# | INFILE-logit.syslog - stdout and stderr for sas call        |
# | INFILE.phi2         - Phi2 estimates from logistic regr.    |
# | INFILE.mlogitdat    - Data used for mult. logist. regr.     |
# | INFILE-mlogit.sas   - Mult. logist. regr. sas file          |
# | INFILE-mlogit.lst   - Mult. logist. regr. sas lst file      |
# | INFILE-mlogit.log   - Mult. logist. regr. sas log file      |
# | INFILE-mlogit.pred  - Pred. values from mult. logist. regr. |
# | INFILE-mlogit.alpha - Alpha estimates & covariance matrix   | 
# | INFILE-mlogit.syslog- stdout and stderr for sas call        |
# | INFILE.phi1         - Phi1 estimates from mult. log. regr.  |
# | INFILE.xi           - Xi estimates                          |
# | INFILE.I3Inv        - The matrix Inverse(I3)                |
# | INFILE.dAdb         - The matrix d(A1,A2) / dBeta           |
# | INFILE.dAda         - The matrix d(A1,A2) / dAlpha          |
# | INFILE.dAdx         - The matrix d(A1,A2) / dXi             |
# | INFILE-iml.sas      - cov(A1,A2) sas iml program            |
# | INFILE-iml.lst      - cov(A1,A2) sas iml lst file           |
# | INFILE-iml.log      - cov(A1,A2) sas iml log file           |
# | INFILE-iml.cov      - cov(A1,A2) results from sas iml       |
# | INFILE-iml.syslog   - stdout and stderr for sas call        |
# | INFILE-pnorm.sas    - p-val and CI sas file                 |
# | INFILE-pnorm.log    - p-val and CI sas log file             |
# | INFILE-pnorm.lst    - p-val and CI sas lst file             |
# | INFILE-pnorm.syslog - stdout and stderr for sas call        |
# |                                                             |
# |Author: Rick Higgs						|
# |Created: 20 July 97						|
# |								|
# |     Date                   Modification			|
# |  ----------     ----------------------------------		|
# |  27 Mar 98      Added option to jacknife variance.          | 
# |  03 Apr 98      Added error check for > 0 obs for every     |
# |                 (T1,T2) combination.                        |
# |  17 Apr 98      Added contingency table input option.       |
# |  17 Apr 98      Modified LOGISTIC and CATMOD to work from   |
# |                 contingency table input.                    |
# |  18 Jun 98      Added option to compute and output          |
# |                 estimated cell counts and probabilities     |
# |								|
# |-------------------------------------------------------------|

use Getopt::Long;

$ret = &GetOptions('i=s', \$inFile,
                   'j',   \$jacknife,
                   't',   \$tableInput,
                   'v',   \$verbose,
                   'c=f', \$confidencePct,
                   'e=s', \$cellFile,
                   'o=s', \$outFile);

if ($confidencePct eq ""){
  $confidencePct = 95.0;
}
if (($confidencePct < 0.0) || ($confidencePct >= 100.0)){
  print_usage();
  die;
}

if ($confidencePct < 1.0){
  print STDERR "Warning: Confidence level should be specified as a percent (e.g. 95 and not 0.95)\n";
}

if ($inFile eq ""){
  print_usage();
  die;
}

if ($outFile eq ""){
  $outFile = $inFile . ".comproc";
}

# ------------------------------------------
# Determine if first row of file is a header
# ------------------------------------------
$firstObs = 1;
open(INFILE, $inFile) or die "Can't open $inFile\n";
$line = <INFILE>;
@tokens = split(" ", $line);
for ($i=1; $i<@tokens; $i++){
  if (($tokens[$i] =~ /^[a-z]*$/) || ($tokens[$i] =~ /^[A-Z]*$/)){
    $firstObs = 2;
  }
}
close(INFILE);

# ----------------------------
# Read in the observation data
# ----------------------------
if ($tableInput){
  open (INFILE, $inFile) or die "Can't open $inFile\n";
  if ($firstObs == 2){
    $line = <INFILE>;
  }
  while($line = <INFILE>){
    ($t1Str, $t2Str, $xStr, $dStr, $cellSum) = split(" ", $line);

    if ($t1Str > $K){
      $K = $t1Str;
    }
    if ($t2Str > $K){
      $K = $t2Str;
    }
    if ($xStr > $I){
      $I = $xStr;
    }

    $numObs += $cellSum;
    
    $n[$xStr] += $cellSum;

    $t1t2Key = $t1Str . ":" . $t2Str;
    $t1t2{$t1t2Key} += $cellSum;

    $cellKey = $t1Str . ":" . $t2Str . ":" . $xStr . ":" . $dStr;
    $obsCellCount{$cellKey} += $cellSum;

    if ($dStr ne "."){
      $cellKey = $t1Str . ":" . $t2Str . ":" . $xStr . ":" . $dStr;
      $cellCount{$cellKey} += $cellSum;
    }

    $tempMCellKey = $xStr . ":" . $t1Str . ":" . $t2Str;
    $tempMCellCount{$tempMCellKey} += $cellSum;
  }
  close(INFILE);

  foreach (sort keys %tempMCellCount){
    ($xStr, $t1Str, $t2Str) = split(":", $_);
    $y = ($t1Str - 1) * $K + ($t2Str - 1);
    $mCellKey = $xStr . ":" . $y;
    $mCellCount{$mCellKey} += $tempMCellCount{$_};
  }
}
else{
  $numObs = 0;
  $I = 0;
  $K = 0;
  open (INFILE, $inFile) or die "Can't open $inFile\n";
  if ($firstObs == 2){
    $line = <INFILE>;
  }
  while($line = <INFILE>){
    ($name[$numObs], $t1[$numObs], $t2[$numObs], $x[$numObs], $d[$numObs]) = split(" ", $line);
  
    $t1t2Key = $t1[$numObs] . ":" . $t2[$numObs];
    $t1t2{$t1t2Key}++;

    if ($t1[$numObs] > $K){
      $K = $t1[$numObs];
    }
    if ($t2[$numObs] > $K){
      $K = $t2[$numObs];
    }
    if ($x[$numObs] > $I){
      $I = $x[$numObs];
    }

    $numObs++;
  }
  close(INFILE);
}

# ----------------------------------------------------------------
# Check to make sure > 0 observations at every (T1,T2) combination
# ----------------------------------------------------------------
for ($i=1; $i<=$K; $i++){
  for($j=1; $j<=$K; $j++){
    $t1t2Key = $i . ":" . $j;
    if ($t1t2{$t1t2Key} == 0){
      open(OUTFILE, ">$outFile") or die "Can't open $outFile\n";
      print OUTFILE "Comparison of effectiveness of two tests based on the area under ROC curve\n";
      print OUTFILE "in the presence of verification bias.\n";
      print OUTFILE "Analysis of $inFile with $numObs observations, $I covariate patterns, and $K test levels.\n";
      print OUTFILE "comproc.pl: Error - No observations for T1=$i and T2=$j.  Too sparse!\n";
      close(OUTFILE);
      die;    
    }
  }
}

# -----------------------------------------------------------
# Count the number of observations for each covariate pattern
# -----------------------------------------------------------
if (! $tableInput){
  for ($i=0; $i<$numObs; $i++){
    $n[$x[$i]]++;
  }
}

# --------------------------------------
# Output a count table for PROC LOGISTIC
# --------------------------------------
$logitDatFile = $inFile . ".logitdat";
$fileList{$logitDatFile} = 1;

if (! $tableInput){
  for ($i=0; $i<$numObs; $i++){
    $cellKey = $t1[$i] . ":" . $t2[$i] . ":" . $x[$i] . ":" . $d[$i] ;
    $obsCellCount{$cellKey}++;

    if ($d[$i] ne "."){
      $cellKey = $t1[$i] . ":" . $t2[$i] . ":" . $x[$i] . ":" . $d[$i] ;
      $cellCount{$cellKey}++;
    }
  }
}

open (OUTFILE, ">$logitDatFile") or die "Can't open $logitDatFile\n";
  foreach (sort keys %cellCount){
    ($t1Cell, $t2Cell, $xCell, $dCell) = split(":", $_);
    $t1Code = &code_it($t1Cell, $K);
    $t2Code = &code_it($t2Cell, $K);
    $xCode = &code_it($xCell, $I);
    print OUTFILE "$_$t1Code$t2Code$xCode $dCell $cellCount{$_}\n";
  }
close(OUTFILE);

# ----------------------
# Call SAS PROC LOGISTIC
# ----------------------
$sasFile  = $inFile . "-logit.sas";
$lstFile  = $inFile . "-logit.lst";
$logFile  = $inFile . "-logit.log";
$predFile  = $inFile . "-logit.pred";
$betaFile = $inFile . "-logit.beta";
$sysLogFile = $inFile . "-logit.syslog";

$fileList{$sasFile} = 1;
$fileList{$lstFile} = 1;
$fileList{$logFile} = 1;
$fileList{$predFile} = 1;
$fileList{$betaFile} = 1;
$fileList{$sysLogFile} = 1;

open (SASFILE, ">$sasFile") or die "Can't open $sasFile\n";
print SASFILE "filename dat \'$logitDatFile\';\n";
print SASFILE "data features;\n";
print SASFILE "length Name \$\ 128;\n";
print SASFILE "infile dat lrecl=1000000 firstobs=1;\n";
print SASFILE "input Name \$";
for ($i=1; $i<$K; $i++){
  $label = "T1" . $i;
  print SASFILE " $label";
}
for ($i=1; $i<$K; $i++){
  $label = "T2" . $i;
  print SASFILE " $label";
}
for ($i=1; $i<$I; $i++){
  $label = "X" . $i;
  print SASFILE " $label";
}
print SASFILE " D COUNT;\n";
print SASFILE "run;\n";

print SASFILE "\nproc logistic data=features descending outest=betas covout;\n";
print SASFILE "freq COUNT;\n";
print SASFILE "model D =";
for ($i=1; $i<$K; $i++){
  $label = "T1" . $i;
  print SASFILE " $label";
}
for ($i=1; $i<$K; $i++){
  $label = "T2" . $i;
  print SASFILE " $label";
}
for ($i=1; $i<$I; $i++){
  $label = "X" . $i;
  print SASFILE " $label";
}
print SASFILE ";\n";
print SASFILE "run;\n";

print SASFILE "\ndata _null_;\n";
print SASFILE "set betas;\n";
print SASFILE "file \'$betaFile\' LRECL=1000000;\n";
print SASFILE "put INTERCEP";
for ($i=1; $i<$K; $i++){
  $label = "T1" . $i;
  print SASFILE " $label";
}
for ($i=1; $i<$K; $i++){
  $label = "T2" . $i;
  print SASFILE " $label";
}
for ($i=1; $i<$I; $i++){
  $label = "X" . $i;
  print SASFILE " $label";
}
print SASFILE ";\n";
print SASFILE "run;\n";
close(SASFILE);

$cmd = "sas $sasFile -unbuflog -noterminal > $sysLogFile 2>&1";
system($cmd);

# -----------------------------
# Parse SAS output to get betas
# -----------------------------
open (BETAFILE, $betaFile) or die "Can't open $betaFile  system($cmd) call returned $?\n";
$line = <BETAFILE>;
@beta = split(" ", $line);
close(BETAFILE);

# -----------------------
# Compute phi2 from betas
# -----------------------
for ($i=1; $i<=$I; $i++){
  for ($j=1; $j<=$K; $j++){
    for ($k=1; $k<=$K; $k++){
      $t1Code = &code_it($j, $K);
      $t2Code = &code_it($k, $K);
      $xCode = &code_it($i, $I);
      $zCode = "1" . " " . $t1Code . " " . $t2Code . " " . $xCode;
      @z = split(" ", $zCode);
      $expSum = 0;
      for($b=0; $b<@beta; $b++){
        $expSum += $z[$b] * $beta[$b]; 
      }
      $phi2[$i][$j][$k] = exp($expSum) / (1 + exp($expSum));
    }
  }
}

# -----------
# Output phi2
# -----------
$phi2File = $inFile . ".phi2";
$fileList{$phi2File} = 1;
open (PHI2FILE, ">$phi2File") or die "Can't open $phi2File\n";
print PHI2FILE "i j k phi_2[i][j][k]\n";
for($i=1; $i<=$I; $i++){
  for($j=1; $j<=$K; $j++){
    for($k=1; $k<=$K; $k++){
      print PHI2FILE "$i $j $k $phi2[$i][$j][$k]\n";
    }
  }
}
close(PHI2FILE);

# ------------------------------------
# Output a count table for PROC CATMOD
# ------------------------------------
$mLogitDatFile = $inFile . ".mlogitdat";
$fileList{$mLogitDatFile} = 1;

if (! $tableInput){
  for ($i=0; $i<$numObs; $i++){
    $y = ($t1[$i] - 1) * $K + ($t2[$i] - 1);
    $mCellKey = $x[$i] . ":" . $y;
    $mCellCount{$mCellKey}++;
  }
}

open (OUTFILE, ">$mLogitDatFile") or die "Can't open $mLogitDatFile\n";
  foreach (sort keys %mCellCount){
    ($xCell, $yCell) = split(":", $_);
    $xCode = &code_it($xCell, $I);
    print OUTFILE "$_$xCode $yCell $mCellCount{$_}\n";
  }
close(OUTFILE);

# --------------------
# Call SAS PROC CATMOD
# --------------------
$sasFile  = $inFile . "-mlogit.sas";
$lstFile  = $inFile . "-mlogit.lst";
$logFile  = $inFile . "-mlogit.log";
$predFile  = $inFile . "-mlogit.pred";
$alphaFile = $inFile . "-mlogit.alpha";
$sysLogFile = $inFile . "-mlogit.syslog";

$fileList{$sasFile} = 1;
$fileList{$lstFile} = 1;
$fileList{$logFile} = 1;
$fileList{$predFile} = 1;
$fileList{$alphaFile} = 1;
$fileList{$sysLogFile} = 1;

open (SASFILE, ">$sasFile") or die "Can't open $sasFile\n";
print SASFILE "filename dat \'$mLogitDatFile\';\n";
print SASFILE "data features;\n";
print SASFILE "length Name \$\ 128;\n";
print SASFILE "infile dat lrecl=1000000 firstobs=1;\n";
print SASFILE "input Name \$";
for ($i=1; $i<$I; $i++){
  $label = "X" . $i;
  print SASFILE " $label";
}
print SASFILE " Y COUNT;\n";
print SASFILE "run;\n";

print SASFILE "\nproc catmod data=features;\n";
print SASFILE "weight COUNT;\n";
print SASFILE "direct";
for ($i=1; $i<$I; $i++){
  $label = "X" . $i;
  print SASFILE " $label";
}
print SASFILE ";\n";
print SASFILE "response logit / out=pred outest=betas;\n";
print SASFILE "model Y =";
for ($i=1; $i<$I; $i++){
  $label = "X" . $i;
  print SASFILE " $label";
}
print SASFILE ";\n";
print SASFILE "run;\n";

print SASFILE "\ndata _null_;\n";
print SASFILE "set pred;\n";
print SASFILE "file \'$predFile\' LRECL=1000000;\n";
print SASFILE "put";
for ($i=1; $i<$I; $i++){
  $label = "X" . $i;
  print SASFILE " $label";
}
print SASFILE " Y _TYPE_ _NUMBER_ _PRED_;\n";
print SASFILE "run;\n";

print SASFILE "\ndata _null_;\n";
print SASFILE "set betas;\n";
print SASFILE "file \'$alphaFile\' LRECL=1000000;\n";
print SASFILE "put";
$numBetas = ($K * $K - 1) * $I;
for ($i=1; $i<=$numBetas; $i++){
  $label = "B" . $i;
  print SASFILE " $label";
}
print SASFILE ";\n";
print SASFILE "run;\n";

close(SASFILE);
$cmd = "sas $sasFile -unbuflog -noterminal > $sysLogFile 2>&1";
system($cmd);

# -------------------------------------------
# Check to make sure multiple logistic worked
# -------------------------------------------
$numLinEqn = ($K * $K) - 1;
$numAlpha = $I * $numLinEqn;
open (ALPHAFILE, $alphaFile) or die "Can't open $alphaFile system($cmd) call returned $?\n";
$lineCtr = -1;
while ($line = <ALPHAFILE>){
  $lineCtr++;
}
if ($lineCtr != $numAlpha){
  if (!$verbose){
    cleanup_files(%fileList);
  } 
  open (OUTFILE, ">$outFile") or die "Can't open $outFile\n";
  print OUTFILE "Comparison of effectiveness of two tests based on the area under ROC curve\n";
  print OUTFILE "in the presence of verification bias.\n";
  printf OUTFILE "Analysis of $inFile with $numObs observations, $I covariate patterns, and $K test levels.\n";
  print OUTFILE "comproc.pl did not complete successfully due to Error in multiple logistic regression\n";
  close(OUTFILE);
  die;
}
close(ALPHAFILE);


# ------------------------------
# Parse SAS output to get alphas 
# ------------------------------
open (ALPHAFILE, $alphaFile) or die "Can't open $alphaFile\n";
$line = <ALPHAFILE>;
@tokens = split(" ", $line);
$i = 0;
for($col=0; $col<$I; $col++){
  for($row=0; $row<$numLinEqn; $row++){
    $alpha[$row][$col] = $tokens[$i];
    $i++;
  }
}
close(ALPHAFILE);

# -----------------------------
# Parse SAS output to get phi_1
# -----------------------------
open (PREDFILE, $predFile) or die "Can't open $predFile\n";
while ($line = <PREDFILE>){
  @tokens = split(" ", $line);
  if ($tokens[$I] eq "PROB"){
    $decodedX = $I;
    for ($i=0; $i<($I - 1); $i++){
      if ($tokens[$i]){
        $decodedX = $i + 1;
      }
    }
    $yVal = $tokens[$I + 1];
    ($t1Val, $t2Val) = decode_y($yVal, $K);
    $phi1[$decodedX][$t1Val][$t2Val] = $tokens[$I + 2];
  }  
}
close(PREDFILE);

# -----------
# Output phi1
# -----------
$phi1File = $inFile . ".phi1";
$fileList{$phi1File} = 1;
open (PHI1FILE, ">$phi1File") or die "Can't open $phi1File\n";
print PHI1FILE "i j k phi_1[i][j][k]\n";
for($i=1; $i<=$I; $i++){
  for($j=1; $j<=$K; $j++){
    for($k=1; $k<=$K; $k++){
      print PHI1FILE "$i $j $k $phi1[$i][$j][$k]\n";
    }
  }
}
close(PHI1FILE);

for($i=1; $i<=$I; $i++){
  $N += $n[$i];
}

for($i=1; $i<=$I; $i++){
  $eta[$i] = $n[$i] / $N;
}

# ----------------------------------------------------------------
# Compute the estimated cell counts and probabilities if specified
# ----------------------------------------------------------------
if ($cellFile ne ""){
  # ----------------------------
  # Compute the P(V=0 | T1,T2,X)
  # ----------------------------
  for($i=1; $i<=$I; $i++){
    for($j=1; $j<=+$K; $j++){
      for($k=1; $k<=$K; $k++){
        $dnom = $obsCellCount{"$j:$k:$i:0"} + $obsCellCount{"$j:$k:$i:1"} + $obsCellCount{"$j:$k:$i:."};
        if ($dnom != 0){
          $v0[$i][$j][$k] = $obsCellCount{"$j:$k:$i:."} / $dnom;
        }
        else{
          $v0[$i][$j][$k] = 0;
        }
      }
    }
  }

  # ---------------------------------------
  # Compute the verified cell probabilities
  # ---------------------------------------
  for ($i=1; $i<=$I; $i++){
    for ($j=1; $j<=$K; $j++){
      for ($k=1; $k<=$K; $k++){
        $str = $j . "-" . $k . "-" . $i . "-0";
        $prob{$str} = (1 - $v0[$i][$j][$k]) * (1 - $phi2[$i][$j][$k]) * $phi1[$i][$j][$k] * $eta[$i];

        $str = $j . "-" . $k . "-" . $i . "-1";
        $prob{$str} = (1 - $v0[$i][$j][$k]) * $phi2[$i][$j][$k] * $phi1[$i][$j][$k] * $eta[$i];
      } #for k
    } #for j
  } #for i

  # -----------------------------------------
  # Compute the unverified cell probabilities
  # -----------------------------------------
  open (CELLFILE, ">$cellFile") or die "Can't open cellFile $cellFile\n";
  print CELLFILE "T1 T2 X D est_prob est_count\n";
  for ($i=1; $i<=$I; $i++){
    for ($j=1; $j<=$K; $j++){
      for ($k=1; $k<=$K; $k++){
        $str = $j . "-" . $k . "-" . $i . "-.";
        $prob{$str} = $v0[$i][$j][$k] * $phi1[$i][$j][$k] * $eta[$i];
        $cellCnt = $prob{$str} * $N;
        print CELLFILE "$j $k $i . $prob{$str} $cellCnt\n";
      } #for k
    } #for j
  } #for i

  # -----------------------------------------------------
  # Output the cell probabilities in easy to read fashion
  # -----------------------------------------------------
  for($i=1; $i<=$I; $i++){
    for($j=1; $j<=$K; $j++){
      for($k=1; $k<=$K; $k++){
        $str0 = $j . "-" . $k . "-" . $i . "-0";
        $str1 = $j . "-" . $k . "-" . $i . "-1";
        $cellCnt0 = $prob{$str0} * $N;
        $cellCnt1 = $prob{$str1} * $N;
        print CELLFILE "$j $k $i 0 $prob{$str0} $cellCnt0\n";
        print CELLFILE "$j $k $i 1 $prob{$str1} $cellCnt1\n";
      } # for k
    } # for j
  } # for i
  close(CELLFILE);
}

# ------------------------------------------
# Compute area under ROC curve for T1 and T2
# ------------------------------------------

# ----------
# Output eta
# ----------
$etaFile = $inFile . ".eta";
$fileList{$etaFile} = 1;
open (ETAFILE, ">$etaFile") or die "Can't open $etaFile\n";
print ETAFILE "i eta[i]\n";
for($i=1; $i<=$I; $i++){
  print ETAFILE "$i $eta[$i]\n";
}
close(ETAFILE);

# Calculate gamma = P(D=1)
# ------------------------
for($i=1; $i<=$I; $i++){
  for($j=1; $j<=$K; $j++){
    for($l=1; $l<=$K; $l++){
      $gamma += $phi2[$i][$j][$l] * $phi1[$i][$j][$l] * $eta[$i];
    } #for l
  } #for j
} #for i

# Calculate phi11
# ---------------
for($j=1; $j<=$K; $j++){
  for($i=1; $i<=$I; $i++){
    for($k=1; $k<=$K; $k++){
      $phi11[$j] += (1 - $phi2[$i][$j][$k]) * $phi1[$i][$j][$k] * $eta[$i];
    }
  }
}

# Calculate phi12
# ---------------
for($j=1; $j<=$K; $j++){
  for($i=1; $i<=$I; $i++){
    for($k=1; $k<=$K; $k++){
      $phi12[$j] += $phi2[$i][$j][$k] * $phi1[$i][$j][$k] * $eta[$i];
    }
  }
}

# Calculate phi21
# ---------------
for($j=1; $j<=$K; $j++){
  for($i=1; $i<=$I; $i++){
    for($k=1; $k<=$K; $k++){
      $phi21[$j] += (1 - $phi2[$i][$k][$j]) * $phi1[$i][$k][$j] * $eta[$i];
    }
  }
}

# Calculate phi22
# ---------------
for($j=1; $j<=$K; $j++){
  for($i=1; $i<=$I; $i++){
    for($k=1; $k<=$K; $k++){
      $phi22[$j] += $phi2[$i][$k][$j] * $phi1[$i][$k][$j] * $eta[$i];
    }
  }
}

# Calculate A1
# ------------
for($j=1; $j<=($K - 1); $j++){
  $innerSum = 0;
  for($l=($j + 1); $l<=$K; $l++){
    $innerSum += $phi12[$l];
  }
  $a1Sum += $phi11[$j] * $innerSum;
}
$a1 = $a1Sum / ($gamma * (1 - $gamma));

# Calculate A2 
# ------------
for($j=1; $j<=($K - 1); $j++){
  $innerSum = 0;
  for($l=($j + 1); $l<=$K; $l++){
    $innerSum += $phi22[$l];
  }
  $a2Sum += $phi21[$j] * $innerSum;
}
$a2 = $a2Sum / ($gamma * (1 - $gamma));

# -----------------------------
# Compute the matrix I3 Inverse
# -----------------------------
$I3file = $inFile . ".I3Inv";
$fileList{$I3file} = 1;
open (I3FILE, ">$I3file") or die "Can't open $I3file\n";
for($row=0; $row<($I - 1); $row++){
  for($col=0; $col<($I - 1); $col++){
    if ($row == $col){
      $I3Inv[$row][$col] = $n[$row + 1] / ($N)**2 - ($n[$row + 1])**2 / ($N)**3;
    }
    else{
      $I3Inv[$row][$col] = -1 * $n[$row + 1] * $n[$col + 1] / ($N)**3;
    } 
    if ($col == 0){
      print I3FILE "$I3Inv[$row][$col]";
    }
    else{
      print I3FILE " $I3Inv[$row][$col]";
    }
  }  #for col
  print I3FILE "\n";
} #for row
close(I3FILE);

# ------------------------
# Compute d(A1,A2) / dBeta
# ------------------------
$numBeta = @beta;
for($i=1; $i<=$I; $i++){
  for($j=1; $j<=$K; $j++){
    for($k=1; $k<=$K; $k++){
      $xCode = &code_it($i, $I);
      $t1Code = &code_it($j, $K);
      $t2Code = &code_it($k, $K);
      $zCode = "1" . " " . $t1Code . " " . $t2Code . " " . $xCode;
      @z = split(" ", $zCode);
      $lin = 0;
      for($b=0; $b<$numBeta; $b++){
        $lin += $z[$b] * $beta[$b];
      }
      $e = exp($lin) / (1 + exp($lin))**2;
      for($b=0; $b<$numBeta; $b++){
        $dPhi2dB[$i][$j][$k][$b] = $e * $z[$b];  
      }
    } #for k
  } #for j
} #for i

for($j=1; $j<=$K; $j++){
  for($b=0; $b<$numBeta; $b++){
    for($i=1; $i<=$I; $i++){
      for($k=1; $k<=$K; $k++){
        $dPhi11dB[$j][$b] += -1 * $dPhi2dB[$i][$j][$k][$b] * $phi1[$i][$j][$k] * $eta[$i];
        $dPhi12dB[$j][$b] += $dPhi2dB[$i][$j][$k][$b] * $phi1[$i][$j][$k] * $eta[$i];
        $dPhi21dB[$j][$b] += -1 * $dPhi2dB[$i][$k][$j][$b] * $phi1[$i][$k][$j] * $eta[$i];
        $dPhi22dB[$j][$b] += $dPhi2dB[$i][$k][$j][$b] * $phi1[$i][$k][$j] * $eta[$i];
      }
    } #for i
  }
} #for j

# Compute dGamma/dBeta
# --------------------
for($b=0; $b<$numBeta; $b++){
  for($j=1; $j<=$K; $j++){
    $dGamdB[$b] += $dPhi11dB[$j][$b] * $phi12[$j] + $phi11[$j] * $dPhi12dB[$j][$b];
  }
}

# Compute dA1/dBeta
# -----------------
for($b=0; $b<$numBeta; $b++){
  $sum1 = 0;
  for($j=1; $j<=($K - 1); $j++){
    for($l=($j + 1); $l<=$K; $l++){
      $sum1 += $dPhi11dB[$j][$b] * $phi12[$l] + $phi11[$j] * $dPhi12dB[$l][$b];
    } # for l
  } #for j  
  $dA1dB[$b] = ($sum1 - $a1 * (1 - 2 * $gamma) * $dGamdB[$b]) / ($gamma * (1 - $gamma));
} #for b

# Compute dA2/dBeta
# -----------------
for($b=0; $b<$numBeta; $b++){
  $sum1 = 0;
  for($j=1; $j<=($K - 1); $j++){
    for($l=($j + 1); $l<=$K; $l++){
      $sum1 += $dPhi21dB[$j][$b] * $phi22[$l] + $phi21[$j] * $dPhi22dB[$l][$b];
    } # for l
  } #for j  
  $dA2dB[$b] = ($sum1 - $a2 * (1 - 2 * $gamma) * $dGamdB[$b]) / ($gamma * (1 - $gamma));
} #for b

$dAdBFile = $inFile . ".dAdB";
$fileList{$dAdBFile} = 1;
open (MATFILE, ">$dAdBFile") or die "Can't open $dAdBFile\n";
  for($b=0; $b<$numBeta; $b++){
    print MATFILE "$dA1dB[$b] $dA2dB[$b]\n";
  }
close(MATFILE);

# -------------------------
# Compute d(A1,A2) / dAlpha
# -------------------------
$maxYVal = $K * $K - 1;
for($i=1; $i<=$I; $i++){
  for($j=1; $j<=$K; $j++){
    for($k=1; $k<=$K; $k++){
      $xCode = &code_it($i, $I);
      $xCode = "1" . " " . $xCode;
      @xVec = split(" ", $xCode);
      $numXParms = @xVec;
      $yCode = ($j - 1) * $K + ($k - 1) + 1;
      ($t1VecVal, $t2VecVal) = decode_y($yCode, $K);
   
      $a = 0;
      for($group=0; $group<=($I -1); $group++){
        for($yVal=1; $yVal<=$maxYVal; $yVal++){
          ($t1Val, $t2Val) = decode_y($yVal, $K);
          if (($yCode != ($maxYVal + 1)) && ($yCode == $yVal)){
            $p1 = $phi1[$i][$t1Val][$t2Val];
            $p2 = $phi1[$i][$t1VecVal][$t2VecVal];
            $dPhi1dA[$i][$j][$k][$a] = $xVec[$group] * ($p2 - $p2 * $p2);
          }
          else{
            $p1 = $phi1[$i][$t1Val][$t2Val];
            $p2 = $phi1[$i][$t1VecVal][$t2VecVal];
            $dPhi1dA[$i][$j][$k][$a] = -1 * $xVec[$group] * $p1 * $p2;
          }
          $a++;
        } #for yVal
      } #for group
    } #for k
  } #for j
} #for i

for($j=1; $j<=$K; $j++){
  for($a=0; $a<$numAlpha; $a++){
    for($i=1; $i<=$I; $i++){
      for($k=1; $k<=$K; $k++){
        $dPhi11dA[$j][$a] += (1 - $phi2[$i][$j][$k]) * $dPhi1dA[$i][$j][$k][$a] * $eta[$i];
        $dPhi12dA[$j][$a] += $phi2[$i][$j][$k] * $dPhi1dA[$i][$j][$k][$a] * $eta[$i];
        $dPhi21dA[$j][$a] += (1 - $phi2[$i][$k][$j]) * $dPhi1dA[$i][$k][$j][$a] * $eta[$i];
        $dPhi22dA[$j][$a] += $phi2[$i][$k][$j] * $dPhi1dA[$i][$k][$j][$a] * $eta[$i];
      }
    } #for i
  }
} #for j


# Compute dGamma/dAlpha
# ---------------------
for($a=0; $a<$numAlpha; $a++){
  for($j=1; $j<=$K; $j++){
    $dGamdA[$a] += $dPhi11dA[$j][$a] * $phi12[$j] + $phi11[$j] * $dPhi12dA[$j][$a];
  }
}

# Compute dA1/dAlpha
# ------------------
for($a=0; $a<$numAlpha; $a++){
  $sum1 = 0;
  for($j=1; $j<=($K - 1); $j++){
    for($l=($j + 1); $l<=$K; $l++){
      $sum1 += $dPhi11dA[$j][$a] * $phi12[$l] + $phi11[$j] * $dPhi12dA[$l][$a];
    } # for l
  } #for j  
  $dA1dA[$a] = ($sum1 - $a1 * (1 - 2 * $gamma) * $dGamdA[$a]) / ($gamma * (1 - $gamma));
} #for a

# Compute dA2/dAlpha
# ------------------
for($a=0; $a<$numAlpha; $a++){
  $sum1 = 0;
  for($j=1; $j<=($K - 1); $j++){
    for($l=($j + 1); $l<=$K; $l++){
      $sum1 += $dPhi21dA[$j][$a] * $phi22[$l] + $phi21[$j] * $dPhi22dA[$l][$a];
    } # for l
  } #for j  
  $dA2dA[$a] = ($sum1 - $a2 * (1 - 2 * $gamma) * $dGamdA[$a]) / ($gamma * (1 - $gamma));
} #for a

$dAdAFile = $inFile . ".dAdA";
$fileList{$dAdAFile} = 1;
open (MATFILE, ">$dAdAFile") or die "Can't open $dAdAFile\n";
  for($a=0; $a<$numAlpha; $a++){
    print MATFILE "$dA1dA[$a] $dA2dA[$a]\n";
  }
close(MATFILE);

# -----------------------
# Compute d(A1,A2) / dEta
# -----------------------
for($j=1; $j<=$K; $j++){
  for($e=1; $e<$I; $e++){
    for($k=1; $k<=$K; $k++){
      $dPhi11dE[$j][$e] += (1 - $phi2[$e][$j][$k]) * $phi1[$e][$j][$k];
      $dPhi12dE[$j][$e] += $phi2[$e][$j][$k] * $phi1[$e][$j][$k];
      $dPhi21dE[$j][$e] += (1 - $phi2[$e][$k][$j]) * $phi1[$e][$k][$j];
      $dPhi22dE[$j][$e] += $phi2[$e][$k][$j] * $phi1[$e][$k][$j];
    } #for k
  } #for e
} #for j

# Compute dGamma/dEta
# -------------------
for($e=1; $e<$I; $e++){
  for($j=1; $j<=$K; $j++){
    $dGamdE[$e] += $dPhi11dE[$j][$e] * $phi12[$j] + $phi11[$j] * $dPhi12dE[$j][$e];
  }
}

# Compute dA1/dEta
# ----------------
for($e=1; $e<$I; $e++){
  $sum1 = 0;
  for($j=1; $j<=($K - 1); $j++){
    for($l=($j + 1); $l<=$K; $l++){
      $sum1 += $dPhi11dE[$j][$e] * $phi12[$l] + $phi11[$j] * $dPhi12dA[$l][$e];
    } # for l
  } #for j  
  $dA1dE[$e] = ($sum1 - $a1 * (1 - 2 * $gamma) * $dGamdE[$e]) / ($gamma * (1 - $gamma));
} #for e

# Compute dA2/dEta
# ----------------
for($e=1; $e<$I; $e++){
  $sum1 = 0;
  for($j=1; $j<=($K - 1); $j++){
    for($l=($j + 1); $l<=$K; $l++){
      $sum1 += $dPhi21dE[$j][$e] * $phi22[$l] + $phi21[$j] * $dPhi22dE[$l][$e];
    } # for l
  } #for j  
  $dA2dE[$e] = ($sum1 - $a2 * (1 - 2 * $gamma) * $dGamdE[$e]) / ($gamma * (1 - $gamma));
} #for a

$dAdEFile = $inFile . ".dAdE";
$fileList{$dAdEFile} = 1;
open (MATFILE, ">$dAdEFile") or die "Can't open $dAdEFile\n";
  for($e=1; $e<$I; $e++){
    print MATFILE "$dA1dE[$e] $dA2dE[$e]\n";
  }
close(MATFILE);

# ----------------------------------------------------
# Run PROC IML to compute the covariance(A1,A2) matrix
# ----------------------------------------------------
$sasImlFile = $inFile . "-iml.sas";
$lstImlFile = $inFile . "-iml.lst";
$logImlFile = $inFile . "-iml.log";
$resImlFile = $inFile . "-iml.cov";
$sysLogFile = $inFile . "-iml.syslog";

$fileList{$sasImlFile} = 1;
$fileList{$lstImlFile} = 1;
$fileList{$logImlFile} = 1;
$fileList{$resImlFile} = 1;
$fileList{$sysLogFile} = 1;

open(SASFILE, ">$sasImlFile") or die "Can't open $sasImlFile\n";
print SASFILE "filename dat \'$dAdAFile\';\n";
print SASFILE "data dada;\n";
print SASFILE "infile dat lrecl=1000000 firstobs=1;\n";
print SASFILE "input X1-X2;\n";
print SASFILE "run;\n";

print SASFILE "\nfilename dat \'$dAdBFile\';\n";
print SASFILE "data dadb;\n";
print SASFILE "infile dat lrecl=1000000 firstobs=1;\n";
print SASFILE "input X1-X2;\n";
print SASFILE "run;\n";

print SASFILE "\nfilename dat \'$dAdEFile\';\n";
print SASFILE "data dade;\n";
print SASFILE "infile dat lrecl=1000000 firstobs=1;\n";
print SASFILE "input X1-X2;\n";
print SASFILE "run;\n";

print SASFILE "\nfilename dat \'$I3file\';\n";
print SASFILE "data i3;\n";
print SASFILE "infile dat lrecl=1000000 firstobs=1;\n";
$label = "X" . ($I - 1);
print SASFILE "input X1-$label;\n";
print SASFILE "run;\n";

print SASFILE "\nfilename dat \'$alphaFile\';\n";
print SASFILE "data alpha;\n";
print SASFILE "infile dat lrecl=1000000 firstobs=2;\n";
$label = "X" . $numAlpha;
print SASFILE "input X1-$label;\n";
print SASFILE "run;\n";

print SASFILE "\nfilename dat \'$betaFile\';\n";
print SASFILE "data beta;\n";
print SASFILE "infile dat lrecl=1000000 firstobs=2;\n";
$label = "X" . $numBeta;
print SASFILE "input X1-$label;\n";
print SASFILE "run;\n";

print SASFILE "proc iml;\n";
print SASFILE "  start;\n";
print SASFILE "  use work.dada;\n";
print SASFILE "  read all var _all_ into dada;\n";
print SASFILE "  use work.dadb;\n";
print SASFILE "  read all var _all_ into dadb;\n";
print SASFILE "  use work.dade;\n";
print SASFILE "  read all var _all_ into dade;\n";
print SASFILE "  use work.i3;\n";
print SASFILE "  read all var _all_ into i3;\n";
print SASFILE "  use work.alpha;\n";
print SASFILE "  read all var _all_ into alpha;\n";
print SASFILE "  use work.beta;\n";
print SASFILE "  read all var _all_ into beta;\n";
print SASFILE "  dadat = dada`;\n";
print SASFILE "  dadbt = dadb`;\n";
print SASFILE "  dadet = dade`;\n";
print SASFILE "  cov = dadat * alpha * dada + dadbt * beta * dadb + dadet * i3 * dade;\n";
print SASFILE "  create work.cov from cov;\n";
print SASFILE "  append from cov;\n";
print SASFILE "  close work.cov;\n";
print SASFILE "finish;\n";
print SASFILE "run main;\n";

print SASFILE "\ndata _null_;\n";
print SASFILE "set work.cov;\n";
print SASFILE "file \'$resImlFile\' lrecl=1000000;\n";
print SASFILE "put col1 col2;\n";
print SASFILE "run;\n";
close(SASFILE);

$cmd = "sas $sasImlFile -unbuflog -noterminal > $sysLogFile 2>&1";
system($cmd);

# Read in the covariance matrix of (A1,A2)
# ----------------------------------------
open (COVFILE, $resImlFile) or die "Can't open $resImlFile system($cmd) call returned $?\n";
$line = <COVFILE>;
($varA1, $covA1A2) = split(" ", $line);
$line = <COVFILE>;
($covA1A2, $varA2) = split(" ", $line);
close(COVFILE);

# Compute jacknife estimates if specified
# ---------------------------------------
if ($jacknife){
  $jackOutFile = $inFile . "-" . $$ . ".jack";
  $fileList{$jackOutFile} = 1;
  if ($tableInput){
    $cmd = "rocjack.pl -t -i $inFile -o $jackOutFile";
  }
  else{
    $cmd = "rocjack.pl -i $inFile -o $jackOutFile";
  }
  system($cmd);

  open(JACKFILE, $jackOutFile) or die "Can't open $jackOutFile\n";
  $line = <JACKFILE>;
  chop($line);
  ($descr, $jackVarA1) = split("=", $line);

  $line = <JACKFILE>;
  chop($line);
  ($descr, $jackVarA2) = split("=", $line);

  $line = <JACKFILE>;
  chop($line);
  ($descr, $jackVarA1A2) = split("=", $line);

  close(JACKFILE);
}

# Compute two-sided p-value and confidence interval
# -------------------------------------------------
$a1Minusa2 = $a1 - $a2;

if ($jacknife){
  $seA1MinusA2 = sqrt($jackVarA1A2);
}
else{
  $seA1MinusA2 = sqrt($varA1 + $varA2 - 2 * $covA1A2);
}

$zScore = $a1Minusa2  / $seA1MinusA2;
$absZScore = abs($zScore);
$ciArea = 1.0 - ((100.0 - $confidencePct) / 200.0);

$sasPNormFile = $inFile . "-pnorm.sas";
$lstPNormFile = $inFile . "-pnorm.lst";
$logPNormFile = $inFile . "-pnorm.log";
$resPNormFile = $inFile . "-pnorm.res";
$sysLogFile = $inFile . "-pnorm.syslog";

$fileList{$sasPNormFile} = 1;
$fileList{$lstPNormFile} = 1;
$fileList{$logPNormFile} = 1;
$fileList{$resPNormFile} = 1;
$fileList{$sysLogFile} = 1;

open (OUTFILE, ">$sasPNormFile") or die "Can't open $sasPNormFile\n";
print OUTFILE "data _null_;\n";
print OUTFILE "  file \'$resPNormFile\';\n";
print OUTFILE "  p=probnorm($absZScore);\n";
print OUTFILE "  q=probit($ciArea);\n";
print OUTFILE "  put p q;\n";
print OUTFILE "run;\n";
close(OUTFILE);

$cmd = "sas $sasPNormFile -unbuflog -noterminal > $sysLogFile 2>&1";
system($cmd);

# Parse output from probnorm and probit
# -------------------------------------
open (RESFILE, $resPNormFile) or die "Can't open $resPNormFile system($cmd) call returned $?\n";
$line = <RESFILE>;
($p1, $q) = split(" ", $line);
close(RESFILE);

# Compute two-sided p-value
# -------------------------
$p = 2 * (1.0 - $p1);

# Compute CI for A1-A2
# --------------------
$lowCI = $a1Minusa2 - $q * $seA1MinusA2;
$hiCI = $a1Minusa2 + $q * $seA1MinusA2;

# Write results to output file
# ----------------------------
if ($jacknife){
  open (OUTFILE, ">$outFile") or die "Can't open $outFile\n";
  print OUTFILE "Comparison of effectiveness of two tests based on the area under ROC curve\n";
  print OUTFILE "in the presence of verification bias.\n";
  printf OUTFILE "Analysis of $inFile with $numObs observations, $I covariate patterns, and $K test levels.\n";
  printf(OUTFILE "A1 = Area under ROC curve for T1 = %f\n", $a1);
  printf(OUTFILE "A2 = Area under ROC curve for T2 = %f\n", $a2);
  printf(OUTFILE "A1-A2 = %f\n", $a1 - $a2);
  printf(OUTFILE "Variance (jacknife) of A1 = %f\n", $jackVarA1);
  printf(OUTFILE "Variance (jacknife) of A2 = %f\n", $jackVarA2);
  printf(OUTFILE "Variance (jacknife) of (A1-A2) = %f\n", $jackVarA1A2);
  printf(OUTFILE "Two-sided p-value = %f\n", $p);
  printf(OUTFILE "%5.2f Percent CI =[%f,%f]\n", $confidencePct, $lowCI, $hiCI);
  close(OUTFILE);
}
else{
  open (OUTFILE, ">$outFile") or die "Can't open $outFile\n";
  print OUTFILE "Comparison of effectiveness of two tests based on the area under ROC curve\n";
  print OUTFILE "in the presence of verification bias.\n";
  printf OUTFILE "Analysis of $inFile with $numObs observations, $I covariate patterns, and $K test levels.\n";
  printf(OUTFILE "A1 = Area under ROC curve for T1 = %f\n", $a1);
  printf(OUTFILE "A2 = Area under ROC curve for T2 = %f\n", $a2);
  printf(OUTFILE "A1-A2 = %f\n", $a1 - $a2);
  printf(OUTFILE "Variance of A1 = %f\n", $varA1);
  printf(OUTFILE "Variance of A2 = %f\n", $varA2);
  printf(OUTFILE "Covariance between A1 and A2 = %f\n", $covA1A2);
  printf(OUTFILE "Two-sided p-value = %f\n", $p);
  printf(OUTFILE "%5.2f Percent CI =[%f,%f]\n", $confidencePct, $lowCI, $hiCI);
  close(OUTFILE);
}

# Remove all temp files if verbose not specified
# ----------------------------------------------
if (!$verbose){
  cleanup_files(%fileList);  
}



sub code_it{
# |-------------------------------------------------------------|
# |code_it: Returns a dummy coding of a test level score	|
# |         or a covariate pattern number.			|
# |								|
# |Author: Rick Higgs						|
# |Created: 20 July 97						|
# |								|
# |     Date                   Modification			|
# |  ----------     ----------------------------------		|
# |								|
# |-------------------------------------------------------------|

  local ($v, $maxV) = @_;
  local ($i, $retVal);

  for ($i=1; $i<$maxV; $i++){
    if ($i == $v){
      $retVal = $retVal . " 1"; 
    }
    else{
      $retVal = $retVal . " 0";
    }
  }

  return $retVal;

} #code_it


sub print_usage{
# |-------------------------------------------------------------|
# |print_usage: Prints out the command line options for		|
# |             invoking comproc.pl				|
# |								|
# |Author: Rick Higgs						|
# |Created: 20 July 97						|
# |								|
# |     Date                   Modification			|
# |  ----------     ----------------------------------		|
# |								|
# |-------------------------------------------------------------|

  print "Usage: comproc.pl -i INFILE <options>\n\n";
  print "Where:   INFILE       is the name of the observation file to process\n";
  print "                      INFILE must be a 5 column file: Name T1 T2 X D\n";
  print "Options: -v           Verbose mode - all temporary files are not removed\n";
  print "         -c CONFLEVEL Where CONFLEVEL is confidence level (Percent) for A1-A2 <95>\n";
  print "         -j           Jacknife mode.  Use jacknife estimate for Var(A1-A2)\n";
  print "         -t           Contingency table input format  INFILE must be in the following format:\n";
  print "                      T1 T2 X D COUNT\n";
  print "         -e CELLFILE  If specified will write estimated cell counts and probabilities to CELLFILE\n";
  print "         -o OUTFILE   Where OUTFILE is name of results output file <INFILE.comproc>\n";
} #print_usage


sub decode_y{
# |-------------------------------------------------------------|
# |decode_y:  Decode a Y value  (Y=1,...) into T1 and T2 parts	|
# |								|
# |Author: Rick Higgs						|
# |Created: 24 July 97						|
# |								|
# |     Date                   Modification			|
# |  ----------     ----------------------------------		|
# |								|
# |-------------------------------------------------------------|
  local ($yVal, $k) = @_;
  local ($t1Val, $t2Val);

  $t1Val = int($yVal / $k) + 1;
  if (int($yVal / $k) == ($yVal / $k)){
    $t1Val--;
  }
  $t2Val = $yVal - ($t1Val - 1) * $k;
  return ($t1Val, $t2Val);
} #decode_y


sub cleanup_files{
# |-------------------------------------------------------------|
# |cleanup_files:  Remove all temporary files used by comproc.pl|
# |								|
# |Author: Rick Higgs						|
# |Created: 19 Aug 97						|
# |								|
# |     Date                   Modification			|
# |  ----------     ----------------------------------		|
# |								|
# |-------------------------------------------------------------|
  local (%files) = @_;

  foreach (%files){
    if (-e $_){
      `rm $_`;
    }
  }

} #cleanup_files
